/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.7 $
 *********************************************************************/

#include "BmpCodec.h"
#include <io/IO.h>

using namespace util::io;
using namespace util;
using namespace std;

namespace prapi
{
	void ColorBmpCodec::encodeMatrix(ostream& out, const Matrix<RGBColor<> >& mat) throw (MatrixException&, IOException&) 
	{
		//cerr << "1\n";
		BitmapFileHeader fileHeader;
		BitmapInfoHeader infoHeader;
		
		int rowLength = mat.getColumns() * 3;
		int remainder = rowLength & 3;
		if (remainder)
			rowLength += 4 - remainder;
		//cerr << "Encoding BMP, rowlength = " << rowLength << "\n";
		fileHeader.bfType = 0x424d;
		fileHeader.bfSize = 54 + rowLength * mat.getRows();
		chOrder(fileHeader.bfSize);
		fileHeader.bfReserved1 = 0;
		fileHeader.bfReserved2 = 0;
		fileHeader.bfOffBits = 54;
		chOrder(fileHeader.bfOffBits);

		infoHeader.biSize = 40;
		chOrder(infoHeader.biSize);
		infoHeader.biWidth =  mat.getColumns();
		chOrder(infoHeader.biWidth);
		infoHeader.biHeight = mat.getRows();
		chOrder(infoHeader.biHeight);
		infoHeader.biPlanes = 1;
		chOrder(infoHeader.biPlanes);
		infoHeader.biBitCount = 24;
		chOrder(infoHeader.biBitCount);
		infoHeader.biCompression = 0;
		infoHeader.biSizeImage = 0;
		infoHeader.biXPelsPerMeter = 6000;
		chOrder(infoHeader.biXPelsPerMeter);
		infoHeader.biYPelsPerMeter = 6000;
		chOrder(infoHeader.biYPelsPerMeter);
		infoHeader.biClrUsed = 0;
		infoHeader.biClrImportant = 0;

		unsigned char* data = new unsigned char[mat.getColumns()*mat.getRows()*3];
		int index = 0, r, c;
		AllItems(r,c,mat)
			{
				RGBColor<>  clr = mat(r,c);
				data[index++] = (unsigned char)clr.getBlue();
				data[index++] = (unsigned char)clr.getGreen();
				data[index++] = (unsigned char)clr.getRed();
			}

		//cerr << "Data read from image.\n";

		rowLength = mat.getColumns() * 3;
		char zeros[4] = {0,0,0,0};

		//out.write((char*)fileHeader,14); //sizeof(BitmapFileHeader));
		out.write((char*)&fileHeader.bfType,2);
		out.write((char*)&fileHeader.bfSize,4);
		out.write((char*)&fileHeader.bfReserved1,2);
		out.write((char*)&fileHeader.bfReserved2,2);
		out.write((char*)&fileHeader.bfOffBits,4);

		//cerr << "Wrote file header.\n";

		out.write((char*)&infoHeader,sizeof(BitmapInfoHeader));
		//cerr << "Wrote info header.\n";
		for (int i=mat.getRows()-1;i>=0;i--)
			{
				out.write((char*)(data+i*rowLength),rowLength);
				if (remainder)
					out.write(zeros,4-remainder);
			}
		delete[] data;
	}


	void ColorBmpCodec::decodeMatrix(istream& in, Matrix< RGBColor<> >& mat) throw (MatrixException&, IOException&)
	{
		BitmapFileHeader fileHeader;
		BitmapInfoHeader infoHeader;

		//in.read((char*)fileHeader,14); //sizeof(BitmapFileHeader));
		inRead(in,(char*)&fileHeader.bfType,2);
		inRead(in,(char*)&fileHeader.bfSize,4);
		inRead(in,(char*)&fileHeader.bfReserved1,2);
		inRead(in,(char*)&fileHeader.bfReserved2,2);
		inRead(in,(char*)&fileHeader.bfOffBits,4);

		inRead(in,(char*)&infoHeader,sizeof(BitmapInfoHeader));


		unsigned short planes = infoHeader.biPlanes;
		unsigned short depth = infoHeader.biBitCount;

		chOrder(infoHeader.biHeight);
		chOrder(infoHeader.biWidth);
		chOrder(planes);
		chOrder(depth);

		if (fileHeader.bfType != 0x424d)
			throw MatrixException("DecodeMatrix: Not a BMP image.");
		if (planes != 1 ||
				depth != 24 ||
				infoHeader.biCompression != 0)
			{
				//cerr << "Planes: " << planes << ", depth: " << depth << ", compression: " << infoHeader.biCompression << "\n";
				throw MatrixException("DecodeMatrix: Unsupported image type");
			}

		int rowLength = infoHeader.biWidth * 3;
		int remainder = rowLength & 3;
		int size = infoHeader.biHeight * rowLength;

		unsigned char* data = new unsigned char[size];
		try
			{
				for (int i=infoHeader.biHeight;i--;)
					{
						inRead(in,(char*)(data+i*rowLength),rowLength);
						if (remainder)
							{
								in.ignore(4 - remainder);
								if (!in)
									throw IOException("DecodeMatrix: Corrupted image data");
							}
					}

				mat.setSize(infoHeader.biHeight,infoHeader.biWidth);
				RGBColor<> * matData = mat.getData();
				for (int dataIndex=0,index=0;dataIndex<infoHeader.biHeight*infoHeader.biWidth;dataIndex++,index+=3,matData++)
					{
						RGBColor<> clr(data[index+2],data[index+1],data[index]);
						*matData = clr;
					}
			}
		catch (...)
			{
				delete[] data;
				throw;
			}
		delete[] data;
	}

	unsigned short ColorBmpCodec::chOrder(unsigned short& val)
	{
		unsigned short tmp = val;
		val = (tmp & 0xff00) >> 8;
		val |= (tmp & 0x00ff) << 8;
		return val;
	}

	unsigned ColorBmpCodec::chOrder(unsigned& val)
	{
		unsigned tmp = val;
		val = (tmp & 0xff000000) >> 24;
		val |= (tmp & 0x00ff0000) >> 8;	
		val |= (tmp & 0x0000ff00) << 8;	
		val |= (tmp & 0x000000ff) << 24;
		return val;
	}

	int ColorBmpCodec::chOrder(int& val)
	{
		int tmp = val;
		val = (tmp & 0xff000000) >> 24;
		val |= (tmp & 0x00ff0000) >> 8;	
		val |= (tmp & 0x0000ff00) << 8;	
		val |= (tmp & 0x000000ff) << 24;
		return val;
	}
}
