/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.8 $
 *********************************************************************/

#ifndef _BINARYMORPHOLOGY_H
#define _BINARYMORPHOLOGY_H

#include <Matrix.h>
#include "../ImageTransform.h"
#include "../ConvolutionMask.h"

namespace prapi { namespace binary {

	/**
	 * The BinaryException.
	 **/
	class BinaryException : public ImageTransformException 
	{
	public:
		BinaryException(std::string message) : ImageTransformException(message) {}
	};
	
	/**
	 * A wrapper class for different types of Binary Morphology
	 * operations.<br>
	 *
	 * Binary morphology operations are used like this:<br>
	 *
	 * <pre>
	 * // make first the mask data 
	 * bool maskData[] = {false,true,true,true,true,true};
	 * // Make new ConvolutionMask witch you give the size and the data
	 * ConvolutionMask&lt;bool&gt; mask(2,3,maskData);
	 *
	 * // If the "operation" point of your ConvolutionMask differs
	 * // from the center point of ConvolutionMask the you have to
	 * // set it.
	 * Point&lt;int&gt; point(0,0);
	 * mask.setOrigin(point);
	 *
	 * // Then if you want to set the Border handling fuction
	 * // you must give the second parameter the handling method.
	 * // (first parameter is the mask).
	 * // now you have operator which wants bool Matrix and gives
	 * // bool Matrix.
	 * Dilation&lt;&gt; test(mask,BORDER_PAD_WITH_ZEROS);
	 *
	 * // And then the using of Binary Morphology.
	 * // 
	 * Matrix&lt;bool&gt; mat(..);
	 *
	 * // then use the Operator to Matrix
	 * Matrix&lt;bool&gt; result(mat.getTransformedImage(mat));
	 *
	 * // And the operation have been made.
	 * </pre>
	 **/
  template <class T=bool, class U=bool> class BinaryMorphology: public ImageTransform<U,T>
	{
	public:
		/**
		 * Get the ConvolutionMask.
		 **/
		virtual ConvolutionMask<T> getConvolutionMask() const = 0;
		/**
		 * Set the ConvolutionMask.
		 **/
		virtual void setConvolutionMask(const ConvolutionMask<T>& mask) = 0;
	};
	
	/**
	 * Makes the Binary Opening for given Matrix.
	 **/
	template <class T=bool, class U=bool> class Opening : public BinaryMorphology<T,U>
	{
	public:
		/**
		 * Constructor for Opening.
		 *
		 * @param mask Convolution mask which wanted to use Opening.
		 * @param borderAction how to handel borders.
		 **/
		Opening(ConvolutionMask<T>& mask, BorderAction borderAction=BORDER_PAD_WITH_ZEROS):
			_ConvolutionMask(mask), _BorderAction(borderAction){}

		/**
		 * Makes the opening for the matrix.
		 *
		 * @param mat The matrix wanted to open.
		 **/
		Matrix<U> getTransformedImage(const Matrix<T>& mat) throw (ImageTransformException&);

		ConvolutionMask<T> getConvolutionMask() const {return _ConvolutionMask;}
		void setConvolutionMask(const ConvolutionMask<T>& mask){_ConvolutionMask = mask;}

	private:
		ConvolutionMask<T> _ConvolutionMask;
		BorderAction _BorderAction;
	};
	
	/**
	 * Makes the Binary Closing for given Matrix.
	 **/
	template <class T=bool, class U=bool> class Closing : public BinaryMorphology<T,U>
	{
	public:
		/**
		 * Constructor for Closing.
		 *
		 * @param mask Convolution mask which wanted to use in Closing.
		 * @param borderAction how to handel borders.
		 **/
		Closing(ConvolutionMask<T>& mask, BorderAction borderAction=BORDER_PAD_WITH_ZEROS):
			_ConvolutionMask(mask), _BorderAction(borderAction){}

		/**
		 * Makes the closing for the matrix.
		 *
		 * @param mat The matrix wanted to close.
		 **/
		Matrix<U> getTransformedImage(const Matrix<T>& mat) throw (ImageTransformException&);
		
		ConvolutionMask<T> getConvolutionMask() const {return _ConvolutionMask;}
		void setConvolutionMask(const ConvolutionMask<T>& mask) {_ConvolutionMask = mask;}

	private:
		ConvolutionMask<T> _ConvolutionMask;
		BorderAction _BorderAction;
	};

	/**
	 * Makes the Binary Dilation for given Matrix.
	 **/
	template <class T=bool, class U=bool> class Dilation : public BinaryMorphology<T,U>
	{
	public:
		/**
		 * Constructor for Dilation.
		 *
		 * @param mask Convolution mask which wanted to use in Dilation.
		 * @param borderAction how to handel borders.
		 **/
		Dilation(ConvolutionMask<T>& mask, BorderAction borderAction=BORDER_PAD_WITH_ZEROS):
			_ConvolutionMask(mask), _BorderAction(borderAction){}

		/**
		 * Makes the dilation for the matrix.
		 *
		 * @param mat The matrix wanted to dilate.
		 **/
		Matrix<U> getTransformedImage(const Matrix<T>& mat) throw (ImageTransformException&);
		
		ConvolutionMask<T> getConvolutionMask() const {return _ConvolutionMask;}
		void setConvolutionMask(const ConvolutionMask<T>& mask){_ConvolutionMask = mask;}

	private:
		ConvolutionMask<T> _ConvolutionMask;
		BorderAction _BorderAction;
	};
	
	template <class T, class U> Matrix<U> Dilation<T,U>::getTransformedImage(const Matrix<T>& mat)
		throw (ImageTransformException&)
	{
		int matRows = mat.getRows();
		int matColumns = mat.getColumns();
		int maskHeight = _ConvolutionMask.getRows();
		int maskWidth = _ConvolutionMask.getColumns();
		int maskY = _ConvolutionMask.getOrigin().y;
		int maskX = _ConvolutionMask.getOrigin().x;

		if (maskWidth > matColumns || maskHeight > matRows)
			throw BinaryException("Dilation<T,U>::getTransformedImage(const Matrix<T>&): Mask size exceeds matrix size.");
		if (maskX >= maskWidth || maskY >= maskHeight || maskX < 0 || maskY < 0)
			throw BinaryException("Dilation<T,U>::getTransformedImage(const Matrix<T>&): The origin of the convolution mask is outside of the mask.");
		
		Matrix<U> result(matRows, matColumns);

		if (_BorderAction == BORDER_CROP)
			{
				result.setSize(matRows-maskHeight+1,matColumns-maskWidth+1);
				// start to make the Dilation for Matrix
				// first set the rows and columns to start from pixels which are
				// equal to Mask Origin.
				for(int rows=maskY; rows<(matRows-maskHeight+maskY+1); rows++)
					{
						for(int cols=maskX; cols<(matColumns-maskWidth+maskX+1); cols++)
							{ // chek if there are something at the place
								if(mat(rows,cols) != T(0))
									{ // then add the new values in the result Matrix
										for(int r=(rows-maskY),maskR=0; maskR<maskHeight ;r++,maskR++)
											for(int c=(cols-maskX),maskC=0; maskC<maskWidth;c++,maskC++)
												// typecast the result of & operator to new Matrix
												if(result(r,c) == U(0))result(r,c) = U(_ConvolutionMask(maskR,maskC));
									}//if
							}//for inner
					}//for outer
			}// if BORDER_CROP
		
		else
			{
				Matrix<T> tmp(matRows+maskHeight-1,matColumns+maskWidth-1);
				result.setSize(matRows,matColumns);

				// set the old piksel from the old matrix at their place
				for(int rows=maskY,r=0; rows<(matRows-maskHeight+maskY+1); rows++,r++)
						for(int cols=maskX,c=0; cols<(matColumns-maskWidth+maskX+1); cols++,c++)
							tmp(rows,cols) = mat(r,c);
				
				switch (_BorderAction)
					{
					case BORDER_REFLECT:
						for (int c=0;c<matColumns;c++)
							for (int r=0;r<maskY;r++)
								tmp(maskY-r-1,c+maskX) = tmp(maskY+r,c+maskX);
						
						for (int c=0;c<matColumns;c++)
							for (int r=0;r<maskHeight-maskY-1;r++)
								tmp(maskY+matRows+r,c+maskX) = tmp(maskY+matRows-r-1,c+maskX);
						
						for (int r=0;r<matRows+maskHeight-1;r++)
							for (int c=0;c<maskX;c++)
								tmp(r,maskX-c-1) = tmp(r,maskX+c);
						
						for (int r=0;r<matRows+maskHeight-1;r++)
							for (int c=0;c<maskX;c++)
								tmp(r,maskX+matColumns+c) = tmp(r,maskX+matColumns-c-1);
						
						break;

					case BORDER_CIRCULAR:
						for (int c=0;c<matColumns;c++)
							for (int r=0;r<maskY;r++)
								tmp(r,c+maskX) = tmp(r+matRows,c+maskX);

						for (int c=0;c<matColumns;c++)
							for (int r=0;r<maskHeight-maskY-1;r++)
								tmp(maskY+matRows+r,c+maskX) = tmp(maskY+r,c+maskX);

						for (int r=0;r<matRows+maskHeight-1;r++)
							for (int c=0;c<maskX;c++)
								tmp(r,c) = tmp(r,matColumns+c);
						
						for (int r=0;r<matRows+maskHeight-1;r++)
							for (int c=0;c<maskX;c++)
								tmp(r,maskX+matColumns+c) = tmp(r,maskX+c);
						
						break;

					default:
						break;
					}//switch
				
				// now go throw the whole new matrix (tmp) with new border
				// the funtion is the same as in BORDER_CROP
				for(int rows=maskY; rows<(tmp.getRows()-maskHeight+maskY+1); rows++)
					{
						for(int cols=maskX; cols<(tmp.getColumns()-maskWidth+maskX+1); cols++)
							{ // chek if there are something at the place
								if(tmp(rows,cols) != T(0))
									{ // then add the new values in the result Matrix
										for(int r=(rows-maskY),maskR=0; maskR<maskHeight ;r++,maskR++)
											for(int c=(cols-maskX),maskC=0; maskC<maskWidth;c++,maskC++)
												// typecast the result of & operator to new Matrix
												if(result(r,c) == U(0))result(r,c) = U(_ConvolutionMask(maskR,maskC));				
									}//if
							}//for inner
					}//for outer
			}//else
			
		return result;
	}

	/**
	 * Makes the Binary Erosion for given Matrix.
	 **/
	template <class T=bool, class U=bool> class Erosion : public BinaryMorphology<T,U>
	{
	public:
		/**
		 * Constructor for Erosion.
		 *
		 * @param mask Convolution mask which wanted to use in Erosion.
		 * @param borderAction how to handel borders.
		 **/
		Erosion(ConvolutionMask<T>& mask, BorderAction borderAction=BORDER_PAD_WITH_ZEROS):
			_ConvolutionMask(mask), _BorderAction(borderAction){}

		/**
		 * Makes the erosion for the matrix.
		 *
		 * @param mat The matrix wanted to erose.
		 **/
		Matrix<U> getTransformedImage(const Matrix<T>& mat) throw (ImageTransformException&);
		
		ConvolutionMask<T> getConvolutionMask() const {return _ConvolutionMask;}
		void setConvolutionMask(const ConvolutionMask<T>& mask){_ConvolutionMask = mask;}

	private:
		ConvolutionMask<T> _ConvolutionMask;
		BorderAction _BorderAction;
	};
	
	template <class T, class U> Matrix<U> Erosion<T,U>::getTransformedImage(const Matrix<T>& mat)
		throw (ImageTransformException&)
	{
		int matRows = mat.getRows();
		int matColumns = mat.getColumns();
		int maskHeight = _ConvolutionMask.getRows();
		int maskWidth = _ConvolutionMask.getColumns();
		int maskY = _ConvolutionMask.getOrigin().y;
		int maskX = _ConvolutionMask.getOrigin().x;

		if (maskWidth > matColumns || maskHeight > matRows)
			throw BinaryException("Erosion<T,U>::getTransformedImage(const Matrix<T>&): Mask size exceeds matrix size.");
		if (maskX >= maskWidth || maskY >= maskHeight || maskX < 0 || maskY < 0)
			throw BinaryException("Erosion<T,U>::getTransformedImage(const Matrix<T>&): The origin of the convolution mask is outside of the mask.");
		
		Matrix<U> result(matRows, matColumns);

		if (_BorderAction == BORDER_CROP)
			{
				result.setSize(matRows-maskHeight+1,matColumns-maskWidth+1);
				//result = U(0); // set all pixels first to zero 
				// start to make the Erosion for Matrix
				// first set the rows and columns to start from pixels which are
				// equal to Mask Origin.
				for(int rows=maskY; rows<(matRows-maskHeight+maskY+1); rows++)
					{
						for(int cols=maskX; cols<(matColumns-maskWidth+maskX+1); cols++)
							{ // chek if there are something at the place
								for(int r=(rows-maskY),maskR=0; maskR<maskHeight ;r++,maskR++)
									for(int c=(cols-maskX),maskC=0; maskC<maskWidth;c++,maskC++)
										// check if the result of operator & is false
										// if it is have to break the loops.
										// NOTE USING GOTO COMMAND.
										if(_ConvolutionMask(maskR,maskC) != T(0) && !(int(_ConvolutionMask(maskR,maskC))&int(mat(r,c))))goto End;
								// if every pixel gave the true answer then the pixel will remain
								if(result(rows-maskY, cols-maskX) == U(0))result(rows-maskY, cols-maskX)= U(1);
								
								// NOTE THIS IS THE PLACE WHERE GOTO GOES
								// it just jumps out of loops
							  End:
								; // this is stypid but needeed
							}//for inner
					}//for outer
			}// if BORDER_CROP
		
		else
			{
				Matrix<T> tmp(matRows+maskHeight-1,matColumns+maskWidth-1);
				result.setSize(matRows,matColumns);

				// set the old piksel from the old matrix at their place
				for(int rows=maskY,r=0; rows<(tmp.getRows()-maskHeight+maskY+1); rows++,r++)
						for(int cols=maskX,c=0; cols<(tmp.getColumns()-maskWidth+maskX+1); cols++,c++)
							tmp(rows,cols) = mat(r,c);
				
				switch (_BorderAction)
					{
					case BORDER_REFLECT:
						for (int c=0;c<matColumns;c++)
							for (int r=0;r<maskY;r++)
								tmp(maskY-r-1,c+maskX) = tmp(maskY+r,c+maskX);
						
						for (int c=0;c<matColumns;c++)
							for (int r=0;r<maskHeight-maskY-1;r++)
								tmp(maskY+matRows+r,c+maskX) = tmp(maskY+matRows-r-1,c+maskX);
						
						for (int r=0;r<matRows+maskHeight-1;r++)
							for (int c=0;c<maskX;c++)
								tmp(r,maskX-c-1) = tmp(r,maskX+c);
						
						for (int r=0;r<matRows+maskHeight-1;r++)
							for (int c=0;c<maskX;c++)
								tmp(r,maskX+matColumns+c) = tmp(r,maskX+matColumns-c-1);
						
						break;

					case BORDER_CIRCULAR:
						for (int c=0;c<matColumns;c++)
							for (int r=0;r<maskY;r++)
								tmp(r,c+maskX) = tmp(r+matRows,c+maskX);

						for (int c=0;c<matColumns;c++)
							for (int r=0;r<maskHeight-maskY-1;r++)
								tmp(maskY+matRows+r,c+maskX) = tmp(maskY+r,c+maskX);

						for (int r=0;r<matRows+maskHeight-1;r++)
							for (int c=0;c<maskX;c++)
								tmp(r,c) = tmp(r,matColumns+c);
						
						for (int r=0;r<matRows+maskHeight-1;r++)
							for (int c=0;c<maskX;c++)
								tmp(r,maskX+matColumns+c) = tmp(r,maskX+c);
						
						break;

					default:
						break;
					}//switch
				
				// now go throw the whole new matrix (tmp) with new border
				// the funtion is the same as in BORDER_CROP
				for(int rows=maskY; rows<(tmp.getRows()-maskHeight+maskY+1); rows++)
					{
						for(int cols=maskX; cols<(tmp.getColumns()-maskWidth+maskX+1); cols++)
							{ // chek if there are something at the place
								for(int r=(rows-maskY),maskR=0; maskR<maskHeight ;r++,maskR++)
									for(int c=(cols-maskX),maskC=0; maskC<maskWidth;c++,maskC++)
												// check if the result of operator & is false
												// if it is have to break the loops.
												// NOTE USING GOTO COMMAND.
												if(_ConvolutionMask(maskR,maskC) != T(0) && !(int(_ConvolutionMask(maskR,maskC))&int(tmp(r,c))))goto End1;
												
								// if every pixel gave the true answer then the pixel will remain
								if(result(rows-maskY, cols-maskX) == U(0))result(rows-maskY, cols-maskX)= U(1);
				
								// NOTE THIS IS THE PLACE WHERE GOTO GOES
								// it just jumps out of loops
							  End1:
								; // this is stypid but needeed
							}//for inner
					}//for outer
			}//else
		
		return result;
	}
	
	template <class T, class U> Matrix<U> Opening<T,U>::getTransformedImage(const Matrix<T>& mat)
		throw (ImageTransformException&)
	{ // make the Erosion first and the the dilation
		Erosion<T,U> erosion(_ConvolutionMask,_BorderAction);
		Dilation<U,U> dilation(_ConvolutionMask,_BorderAction);

		return dilation.getTransformedImage(erosion.getTransformedImage(mat));
	}

	template <class T, class U> Matrix<U> Closing<T,U>::getTransformedImage(const Matrix<T>& mat)
		throw (ImageTransformException&)
	{ // make the Dilation first and the the Erosion
		Dilation<T,U> dilation(_ConvolutionMask,_BorderAction);
		Erosion<U,U> erosion(_ConvolutionMask,_BorderAction);

		return erosion.getTransformedImage(dilation.getTransformedImage(mat));
	}
}}
#endif
