/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.7 $
 *********************************************************************/

#ifndef _LABELING_H
#define _LABELING_H

#include "../ImageTransform.h"
#include "../Image.h"
#include <SortedList.h>
#include <Math.h>
#include <ListUtils.h>

namespace prapi { namespace binary {

	/**
	 * LabelingException.
	 **/
	class LabelingException : public MatrixException 
	{
	public:
		LabelingException(char* message) : MatrixException(message) {}
	};
	
	/**
	 * A wrapper class for different types of labeling operators.
	 * operations.
	 **/	
  template <class T=bool, class U=bool> class Labeling : public ImageTransform<U,T> {};

	/**
	 * Class Conectivity4 allows you to make labeling for 4 conected components.
	 **/
	template <class T=bool, class U=bool> class Conectivity4 : public Labeling<T,U>
	{
	public:
		/**
		 * Constructor of Conectivity4.
		 **/
		Conectivity4(){}
		/**
		 * Makes the labeling for 4 conected componets.
		 *
		 * @param mat The matrix wanted to make labeling.
		 **/
		Matrix<U> getTransformedImage(const Matrix<T>& mat)throw (ImageTransformException&);
	};

	/**
	 * Class Conectivity8 allow you to make labeling for 8 conected components.
	 **/
	template <class T=bool, class U=bool> class Conectivity8 : public Labeling<T,U>
	{
	public:
		/**
		 * Constructor of Conectivity8.
		 **/
		Conectivity8(){}
		/**
		 * Makes the labeling for 8 conected componets.
		 *
		 * @param mat The mat wanted to make labeling
		 **/
		Matrix<U> getTransformedImage(const Matrix<T>& mat)throw (ImageTransformException&);

	private:
		/**
		 * The fuction finds the minimum value of the list.
		 **/
		int findMinimum(int index,List<int>& labels);
	};

	template <class T, class U> Matrix<U> Conectivity4<T,U>::getTransformedImage(const Matrix<T>& mat)
		throw (ImageTransformException&)
	{
		int matColumns = mat.getColumns();
		int matRows = mat.getRows();
		int tmpRows = matRows+1;
		int tmpColumns = matColumns+1;
		// make the label list and set the first member to zero.
		List<int> labels(256);
		labels.addElement(0);
		
		Matrix<int> tmp(tmpRows,tmpColumns);

		for(int r=0,i=1; r<matRows;r++,i++)
			{
				for(int c=0,j=1; c<matColumns;c++,j++)
					{
						if(mat(r,c) != T(0))
							{ // check left
								if(tmp(i,j-1))
									{ //check up
										if(tmp(i-1,j))
											{ // if the labels are same give either left or up
												if(tmp(i,j-1) == tmp(i-1,j))tmp(i,j) = labels[tmp(i,j-1)];
												// if different add the labels same
												else
													{ //if left is smaller than up give the up lefts label
														if(tmp(i,j-1) < tmp(i-1,j))
															{
																labels[tmp(i-1,j)]=tmp(i,j-1);
																tmp(i,j)=labels[tmp(i,j-1)];
															}
														//if up's label is smaller than left's then give left the up's label
														else
															{
																labels[tmp(i,j-1)]=tmp(i-1,j);
																tmp(i,j)=labels[tmp(i-1,j)];
															}
													}
											}
										else tmp(i,j)=labels[tmp(i,j-1)];//give the label from left
									}
								// chek up then
								else if(tmp(i-1,j))tmp(i,j)=labels[tmp(i-1,j)]; // add the label from up
								else
									{
										// add the lists length to next label
										tmp(i,j)=labels.getLength();
										labels.addElement(labels.getLength());
									}
							}//if
					}
			}
		
		// go throw the labels
		int labelIndex = 0;
		for(int i=0;i<labels.getLength();i++)
			{
				if(labels[i] != i)labels[i]=labels[labels[i]];		
				else
					{
						labels[i] = labelIndex;
						labelIndex++;
					}
			}

		Matrix<int> result(matRows,matColumns);
		// and then go throw the whole Matrix again and change the labels.
		for(int i=1,r=0; i<tmpRows;i++,r++)
			for(int j=1,c=0; j< tmpColumns;j++,c++)
				{
					if((tmp(i,j) != U(0)) && (labels[tmp(i,j)] != U(0)))result(r,c)=labels[tmp(i,j)];
					else result(r,c)=tmp(i,j); 
				}
		return Matrix<U>(result);
	}

	
	template <class T, class U> Matrix<U> Conectivity8<T,U>::getTransformedImage(const Matrix<T>& mat)
		throw (ImageTransformException&)
	{
		int matColumns = mat.getColumns();
		int matRows = mat.getRows();
		int tmpRows = matRows+2;
		int tmpColumns = matColumns+2;
		// make the label list and set the first member to zero.
		List<int> labels(256);
		labels.addElement(0);
		
		Matrix<int> tmp(tmpRows,tmpColumns);

		// do not even try to understand what this does
		for(int r=0,i=1; r<matRows;r++,i++)
			{
				for(int c=0,j=1; c<matColumns;c++,j++)
					{
						if(mat(r,c) != T(0))
							{
								SortedList<int> values(4);
								values.addElements(4,findMinimum(tmp(i,j-1),labels),findMinimum(tmp(i-1,j-1),labels),
																		 findMinimum(tmp(i-1,j),labels),findMinimum(tmp(i-1,j+1),labels));
								int min=ListUtils::findFirstNotOf(values,0);
								// if everything were not zeros
								if(min != -1)
									{
										min = values[min];
										labels[tmp(i,j-1)]=min;//check left
										labels[tmp(i-1,j-1)]=min;//check northwest
										labels[tmp(i-1,j)]=min;//check up
										labels[tmp(i-1,j+1)]=min;//check northeast
										tmp(i,j)=min;//set the right value
									}
								else
									{
										// add the lists length to next label 
										tmp(i,j)=labels.getLength();
										labels.addElement(labels.getLength());
									}
							}//if
					}
			}
		
		// go throw the labels
		int labelIndex = 1;
		for(int i=0;i<labels.getLength();i++)
			{
				if(labels[i] != i)labels[i]=labels[labels[i]];		
				else
					{
						labels[i] = labelIndex;
						labelIndex++;
					}
			}
		
		Matrix<int> result(matRows,matColumns);
		// and then go throw the whole Matrix again and change the labels.
		for(int i=1,r=0; r<matRows;i++,r++)
			for(int j=1,c=0; c< matColumns;j++,c++)
				{
					if((tmp(i,j) != U(0)) && (labels[tmp(i,j)] != U(0)))result(r,c)=labels[tmp(i,j)];
					else result(r,c)=tmp(i,j); 
		 		}

		return Matrix<U>(result);
	}
	
	template <class T, class U> int Conectivity8<T,U>::findMinimum(int index,List<int>& labels)
	{
		int tmp=labels[index];
		int before=index;
		while(tmp < before && before > 1 && tmp >= 1)
			{
				before=tmp;
				tmp=labels[tmp];
			}
		return before;
	}
	
}}
#endif
	
