/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.3 $
 *********************************************************************/

#ifndef _OUTEXCLASSIFICATIONENGINE_H
#define _OUTEXCLASSIFICATIONENGINE_H

#include <fstream>
#include <string>

#include <Util.h>
#include <List.h>
#include <Matrix.h>
#include <Math.h>
#include <MatrixCodec.h>

#include "../Sample.h"
#include "../ConfusionMatrix.h"
#include "../RasterCodec.h"
#include "../BmpCodec.h"

#include "OutexResult.h"

namespace prapi { namespace extras {
	
	/**
	 * This class automatically parses and classifies Outex test suites.
	 * It automatically creates sample sets and asks a subclass to
	 * perform the actual classification on each iteration of the test
	 * suite.
	 **/	
	template <class T, class I=std::string, class C=int> class OutexClassificationEngine
	{
	public:
		/**
		 * Construct a classification engine.
		 *
		 * @param precalculate a flag that controls the memory consumption
		 * versus speed trade-off. If set to true (the default), feature
		 * vectors for all images are calculated and stored in the main
		 * memory before classification. If there is not enough memory for
		 * this, the flag can be set to false, in which case a feature
		 * vector is calculated each time an image is accessed. Memory
		 * consumption is kept at minimum, but feature vectors must be
		 * calculated many times for each image.
		 **/
		OutexClassificationEngine(bool precalculate = true) : _bPrecalculate(precalculate), _lstConfusionMatrices(0){}

		/**
		 * Clean up any reserved resources.
		 **/
		virtual ~OutexClassificationEngine(){}

		/**
		 * Set the precalculation flag.
		 *
		 * @see OutexClassificationEngine(bool)
		 **/
		void setPrecalculation(bool precalculate) { _bPrecalculate = precalculate; }

		/**
		 * Get the current value of the precalculation flag.
		 **/
		bool getPrecalculation() { return _pPrecalculate; }
		
		/**
		 * Whenever the engine encounters a gray-scale image, this method
		 * is called. The method should extract the needed features and
		 * return them as a feature vector.
		 *
		 * @param mat a gray-scale image matrix
		 **/
		virtual util::List<T> getFeatureVectors(const util::Matrix<int>& mat) { List<T> lst; return lst; }

		/**
		 * Whenever a color image is encountered, this method is consulted
		 * to extract the features.
		 *
		 * @param mat a color image
		 **/
		virtual util::List<T> getFeatureVectors(const util::Matrix<RGBColor<> >& mat){List<T> lst;return lst;}
		
		/**
		 * Once the engine has constructed the necessary sample sets for
		 * classification, it calls this method. The method can use
		 * whatever classification principle you feel suitable. As a
		 * result, the classification of each sample in the test set
		 * should be performed.
		 *
		 * @param train The train set which is used for training.
		 * @param test The test set which is used for testing.
		 * @param classCount the total number of classes
		 **/
		virtual void classify(util::List<Sample<T,I,C> >& train, util::List<Sample<T,I,C> >& test, int classCount) = 0;


		/**
		 * Once each iteration has been completed, the confusion matrices
		 * from each classification are collected in a list. You can get
		 * the list by calling this method.
		 **/
		util::List<ConfusionMatrix> getConfusionMatrices() {return _lstConfusionMatrices;}
		/**
		 * Get the class names from each classification.
		 **/
		util::List<util::List<std::string> > getClasses() { return _lstClasses; }
		/**
		 * Get the class names at the specified classification round.
		 **/
		util::List<std::string> getClasses(int index) { return _lstClasses[index > _lstClasses.getLength() ? 0 : index]; }
		
		/**
		 * The function evaluateSuite classifies all the problems in the test suite
		 * and returns the result in OutexResult format.
		 *
		 * @param path A path to the root of the test suite.
		 *
		 * @return The result of the classification.
		 **/
		OutexResult evaluateSuite(std::string path) throw (InvalidArgumentException&);

	private:
		/**
		 * function reads the Outex format txt file and returns
		 * the list of strings in file.
		 **/
		static util::List<std::string> readOutexTxtFile(std::string fileName);
		/**
		 * function reads the Outex format classes txt file.
		 * Note that the index of class is the same as the place
		 * in the classes list.
		 **/
		static void readOutexClassesFile(std::string fileName,util::List<std::string>& classes,util::List<int>& cost);
		/**
		 * function reads the Outex format train/test txt file.
		 * Note that the index of class is the same as the place
		 * in the classes list.
		 **/
		static void readOutexTrainTestFile(std::string fileName,util::List<std::string>& names,util::List<int>& classIndexes);

	protected:
		/**
		 * Flag for pre calculation.
		 **/
		bool _bPrecalculate;
		/**
		 * ConfusionMatrixses from the classification
		 **/
		util::List<ConfusionMatrix> _lstConfusionMatrices;
		/**
		 * The class names from the problems.
		 * (the classes from the problem 000 is in place 0 and so on.)
		 **/
		util::List<util::List<std::string> > _lstClasses;
	};

	
	template <class T,class I,class C> OutexResult OutexClassificationEngine<T,I,C>::evaluateSuite(std::string path)
		throw (InvalidArgumentException&)
	{
		//variables
		int i=0,imageCount=0;
		RasterCodec ras;
		ColorBmpCodec bmp;

		// first calculate all features from the given images
		util::List<std::string> images(readOutexTxtFile(path+"images.txt"));
		imageCount=images.getLength();
		util::List<util::List<T> > features(imageCount);

		if(_bPrecalculate) // check if precalculation is wanted
			{
				//decide which codec will be used.
				if(images[0].substr(images[0].length()-3)=="ras")
					for(i=0;i<imageCount;i++)// read the image and calculate features
						features += getFeatureVectors(ras.readFromFile(path+images[i]));
				else // color images
					for(i=0;i<imageCount;i++)// read the image and calculate features
						features += getFeatureVectors(bmp.readFromFile(path+images[i]));
			}
				
		// then start processing the problems.
		util::List<std::string> problems(readOutexTxtFile(path+"problems.txt"));
		util::List<double> errors(problems.getLength()); // make the result list

		for(int pIndex=0;pIndex<problems.getLength();pIndex++)
			{
				// make the train set
				util::List<std::string> names; // the amount of image names in train set
				util::List<int> classIndexes;  // the indexes of image classes
				readOutexTrainTestFile(path+problems[pIndex]+"/train.txt",names,classIndexes);
				int trainSetSize=names.getLength();
				util::List<Sample<T,I,C> >  train(trainSetSize);

				// start making the train sample set
				for(i=0;i<trainSetSize;i++)
					{ 
						Sample<T,I,C> sample(classIndexes[i]); // make new sample
						sample.setIdentifier(I(names[i])); // set the identifier

						if(!_bPrecalculate) // check if precalculation is not used
							{
							
								//decide which codec will be used.
								if(images[0].substr(images[0].length()-3)=="ras")
									sample.setFeatureVector(getFeatureVectors(ras.readFromFile(path+"images/"+names[i])));
								else // color images
									sample.setFeatureVector(getFeatureVectors(bmp.readFromFile(path+"images/"+names[i])));
							}
						else
							{
								// calculate the index of image in the list
								int imageIndex=images.indexOf("images/"+names[i]);
								if(imageIndex == -1)
									throw InvalidArgumentException("OutexClassificationEngine<T,I,C>::evaluateSuite(std::string) : Image "+names[i]+" not found.");
						
								sample.setFeatureVector(features[imageIndex]);
							}
						train += sample;
						
						//cerr << "\rMaking train sample: "<<i;
					}
				//cerr << endl;

				
				// make the test set
				readOutexTrainTestFile(path+problems[pIndex]+"/test.txt",names,classIndexes);
				int testSetSize=names.getLength();
				util::List<Sample<T,I,C> >  test(testSetSize);

				// start making the test sample set
				for(i=0;i<testSetSize;i++)
					{ 
						Sample<T,I,C> sample(classIndexes[i]); // make new sample
						sample.setIdentifier(I(names[i])); // set the identifier

						if(!_bPrecalculate) // check if precalculation is not used
							{
							
								//decide which codec will be used.
								if(images[0].substr(images[0].length()-3)=="ras")
									sample.setFeatureVector(getFeatureVectors(ras.readFromFile(path+"images/"+names[i])));
								else // color images
									sample.setFeatureVector(getFeatureVectors(bmp.readFromFile(path+"images/"+names[i])));
							}
						else
							{
								// calculate the index of image in the list
								int imageIndex=images.indexOf("images/"+names[i]);
								if(imageIndex == -1)
									throw InvalidArgumentException("OutexClassificationEngine<T,I,C>::evaluateSuite(std::string) : Image "+names[i]+" not found.");
						
								sample.setFeatureVector(features[imageIndex]);
							}
						test += sample;
						
						//cerr << "\rMaking test sample: "<<i;
					}
				//cerr << endl;

				//read the class and cost information
				util::List<std::string> classes;
				util::List<int> cost;
				readOutexClassesFile(path+problems[pIndex]+"/classes.txt",classes,cost); //read the class and cost information

				_lstClasses += classes;
				
				// then classify the problem
				int classCount = classes.getLength();
				classify(train,test,classCount);
				
				//Then make the confusion Matrix.
				ConfusionMatrix cm(test,classCount);
				_lstConfusionMatrices += cm; //add to the list.

				//cm.print(cout,classes);
				
				// then calculate the result (remember to use cost functions).
				double score = 0, maxScore = 0;
				for (int r=0;r<classCount;r++)
					{
						int correct=0, incorrect=0;
						for (int c=0;c<classCount;c++)
							{
								if (r == c)correct += cm(r,c);
								else incorrect += cm(r,c);
							}
						
						score += cost[r] * incorrect;
						maxScore += (incorrect+correct) * cost[r];
					}
				errors += (score/maxScore); // add the result to the errors list.

				//cerr << "problem: "<<pIndex<<" procecced"<<endl;
			}
		
		// then make the results
		OutexResult result(1-Math::min(errors),1-Math::max(errors),1-Math::mean(errors),Math::stdev(errors));
		
		return result;
	}

	template <class T,class I,class C>
	util::List<std::string> OutexClassificationEngine<T,I,C>::readOutexTxtFile(std::string fileName)
	{
		// read the first line
		std::string line="";
		std::ifstream file(fileName.c_str());
		file >> line;

		// make the List which length is
		int numberOfProblems=Util::parseInt(line);
		util::List<std::string> result(numberOfProblems);

		// and take the 
		for(int i=0;i<numberOfProblems;i++)
			{
				file >> line;
				result += line;
			}
		file.close();
		
		return result;
	}
	
	template <class T,class I,class C>
	void OutexClassificationEngine<T,I,C>::readOutexClassesFile(std::string fileName,util::List<std::string>& classes,util::List<int>& cost)
	{
		// read the first line
		std::string line="";
		std::ifstream file(fileName.c_str());
		
		file >> line;
		
		// make the List which length is
		int numberOfClasses=Util::parseInt(line);
		classes.setLength(0);
		cost.setLength(0);

		// and take the 
		for(int i=0;i<numberOfClasses;i++)
			{
				file >> line; // read the class name
				classes += line; 
				file >> line; // read the class index
				file >> line; // read the cost
				cost += Util::parseInt(line);
			}
		file.close();
	}

	template <class T,class I,class C>
	void OutexClassificationEngine<T,I,C>::readOutexTrainTestFile(std::string fileName,util::List<std::string>& names,util::List<int>& classIndexes)
	{
		// read the first line
		std::string line="";
		std::ifstream file(fileName.c_str());
		file >> line;

		// make the List which length is
		int numberOfSamples=Util::parseInt(line);
		names.setLength(0);
		classIndexes.setLength(0);

		// and take the 
		for(int i=0;i<numberOfSamples;i++)
			{
				file >> line; // read the sample name
				names += line;
				file >> line; // read the class Index
				classIndexes += Util::parseInt(line);
			}
		file.close();
	}
	
}}
#endif
