/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision:$
 *********************************************************************/

#ifndef _CALBP_H
#define _CALBP_H

#include "LBP.h"

namespace prapi { namespace texture {

	template <class T> class CALBP : public FeatureExtractor<int, util::Matrix<T> >
	{
	public:
		CALBP(util::List<unsigned int> predicates, unsigned int samples=8);
		~CALBP();
		
		util::List<int> getFeatureVector(const util::Matrix<T>& image) throw (FeatureExtractionException&);
	private:
		unsigned int _uiMaxPredicate, _uiSamples;
		util::List<GeneralLBP<T>*> _plstLBP;

		void vote(int val1, int val2, util::List<int>& votes);
	};

	template <class T> CALBP<T>::CALBP(util::List<unsigned int> predicates, unsigned int samples) :
		_uiMaxPredicate(0), _uiSamples(samples)
	{
		for (int i=0; i<predicates.getLength(); i++)
			{
				if (predicates[i] > _uiMaxPredicate)
					_uiMaxPredicate = predicates[i];
				_plstLBP += new GeneralLBP<T>(predicates[i], true, samples);
			}
	}

	template <class T> CALBP<T>::~CALBP()
	{
		for (int i=_plstLBP.getLength(); i--; )
			delete _plstLBP[i];
	}

	template <class T> util::List<int> CALBP<T>::getFeatureVector(const util::Matrix<T>& image)
		throw (FeatureExtractionException&)
	{
		int len = _plstLBP.getLength();

		//Create a list of LBP images for different predicates
		util::List<util::Matrix<int> > images(len);
		for (int i=0; i<len; i++)
			{
				//Make sure each image is of the same size, and that all have
				//corresponding pixels in the same coordinates.
				int diff = _uiMaxPredicate - _plstLBP[i]->getPredicate();
				if (diff)
					images += _plstLBP[i]->getTransformedImage(image(diff, diff, image.getRows()-diff*2, image.getColumns()-diff*2));
				else
					images += _plstLBP[i]->getTransformedImage(image);
			}

		util::List<int> result(256);
		result.setLength(256,0);

		int r,c;
		util::List<int> votes(8);
		votes.setLength(8);

		//Go through all pixels in the image
		AllItems(r,c,images[0])
			{
				votes = 0;
				//Each LBP code is used to determine the CA that could create
				//the next binary pattern given the previous one.
				for (int i=0; i<len-1; i++)
					vote(images[i](r,c), images[i+1](r,c), votes);

				//cerr << "Final votes: " << votes << endl;
				int caCode = 0;
				//Negative vote means 0, positive 1. Shift the sign.
				for (int i=0; i<8; i++)
					caCode |= (votes[i] & 0x80000000) >> (31-i);

				//cerr << "Cacode: " << BitOperation::binary(caCode,8) << " (" << caCode << ")" << endl;
				result[caCode]++;
			}

		return result;
	}

	template <class T> void CALBP<T>::vote(int val1, int val2, util::List<int>& votes)
	{
		//cerr << "Code 1: " << BitOperation::binary(val1,_uiSamples) << ", Code 2: " << BitOperation::binary(val2,_uiSamples) << endl;
		unsigned int value = BitOperation::rol((unsigned int)val1, 1, _uiSamples);

		for (unsigned int i=0; i<_uiSamples; i++)
			{
				votes[value & 7] += (val2 & (1 << i)) ? -1 : 1;

				//cerr << "Bit pattern " << i << ": " << BitOperation::binary(value & 7, 3) << " (" << (value & 7)
				//		 << ") votes for " << ((val2 & (1 << i)) ? 1 : 0) << endl;

				value = BitOperation::ror(value, 1, _uiSamples);
			}
	}
}}
		

#endif
