/*********************************************************************
 * This file is part of the PRAPI library.
 *
 * Copyright (C) 2001 Topi Menp and Jaakko Viertola
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.21 $
 *********************************************************************/

#ifndef _EDGEDETECTOR_H
#define _EDGEDETECTOR_H

#include "../ImageTransform.h"
#include "../ConvolutionMask.h"
#include "../Coordinates.h"
#include "../Image.h"
#include "../dsp/Convolution.h"
#include "Thresholding.h"

#include <MatrixUtils.h>
#include <Matrix.h>
#include <math.h>

namespace prapi { namespace texture {

	/**
	 * The EdgeDetectorException.
	 **/
	class EdgeDetectorException : public ImageTransformException 
	{
	public:
		EdgeDetectorException(std::string message) : ImageTransformException(message) {}
	};

	/**
	 * An enumeration for the getGradient function.
	 * <ul>
	 * <li>GRADIENT_X - The gradient in x direction.</li>
	 * <li>GRADIENT_Y - The gradient in y direction.</li>
	 * </ul>
	 **/
	enum Gradient { GRADIENT_X, GRADIENT_Y };

	/**
	 * An enumeration for already implemented, well-known edge
	 * detection masks. These include:
	 * 
	 * <ul>
	 * <li>SOBEL - Sobel in a 3x3 neighborhood</li>
	 * <li>PREWITT - Prewitt in a 3x3 neighborhood</li>
	 * <li>ROBERTS - Roberts in a 2x2 neighborhood</li>
	 * </li>
	 **/
	enum EdgeMask { SOBEL, PREWITT, ROBERTS };
	
	/**
	 * EdgeDectector includes th functions how to use edgeDetectors. Note that EdgeDetector
	 * uses Matrix<Point<T> > type of matrices where the gradient matrices will be placed. 
	 **/
  template <class T, class U=bool> class EdgeDetector: public ImageTransform<U,T>
	{
	public:
		/**
		 * Constructor of EdgeDetector.
		 *
		 * @param borderAction for the border handling. 
		 **/
		EdgeDetector(int radius, BorderAction borderAction = BORDER_REFLECT): _BorderAction(borderAction),_iradius(radius){}
		/**
		 * Destructor of EdgeDetector.
		 **/
		~EdgeDetector(){}
		/**
		 * Function suppresNonMaxima finds the maximum point of the matrix
		 * at the direction of angele matrix.
		 *
		 * @param matMagnitude magnitude matrix which wanted to find maximums.
		 * @param matAngle angle matrix of the magnitude matrix.
		 * @param radius the length how far the maximum will be looked.
		 * @param extendType How to handel borders.
		 **/
		static util::Matrix<T> suppressNonMaxima(const util::Matrix<T>& matMagnitude, const util::Matrix<T>& matAngle, int radius,
																						 util::MatrixUtils::ExtendType extendType = util::MatrixUtils::EXTEND_ZEROPAD)
			throw (EdgeDetectorException&);

		/**
		 * Get the angle matrix from the gradient matrixs X and Y.
		 * NOTE all angles are posite so they are at range [0,2PI].
		 *
		 * @param mat Gradient matrix which icludes X and Y gradients.
		 **/
		static util::Matrix<double> getAngle(const util::Matrix<Point<T> >& mat)
			throw (EdgeDetectorException&);

		/**
		 * Get the magnitude matrix from the gradient matrices X and Y.
		 *
		 * @param mat Gradient matrix which includes X and Y gradients.
		 **/
		static util::Matrix<double> getMagnitude(const util::Matrix<Point<T> >& mat)
			throw (EdgeDetectorException&);
		
		/**
		 * Get either of the gradient coordinate layers from a two-layer
		 * gradient image.
		 *
		 * @param mat The matrix where you want to take the gradient util::Matrix out.
		 * @param gradient The gradient which you want to separate (X or Y).
		 **/
		static util::Matrix<T> getGradient(const util::Matrix<Point<T> >& mat, Gradient gradient)
			throw (EdgeDetectorException&);

		/**
		 * Set a gradient coordinate layer to a value. Using this method
		 * it is possible to alter gradient coordinates separately.
		 *
		 * @param matGradient The matrix which the gradien will be added.
		 * @param mat The matrix which will be added.
		 * @param gradient Which gradient you want to add in here (X or Y).
		 **/
		static void setGradient(util::Matrix<Point<T> >& matGradient, const util::Matrix<T>& mat, Gradient gradient)
			throw (EdgeDetectorException&);
		
	protected:
		/**
		 * BorderAction is for the border handling.
		 **/
		BorderAction _BorderAction;
		/**
		 * _iRadius is used in suppresNonMaxima to tell how far the
		 * suppres will be calculated.
		 **/
		int _iradius;
		
	private:
		/**
		 * The angles used in suppresNonMaxima.
		 **/
		static const double WEST = 7*(M_PI_4/2);
		static const double NORTH = 3*(M_PI_4/2);
		static const double NORTH_EAST = (M_PI_4/2);
		static const double NORTH_WEST = 5*(M_PI_4/2);
		static const double SOUTH = 11*(M_PI_4/2);
		static const double SOUTH_WEST = 9*(M_PI_4/2);
		static const double SOUTH_EAST = 13*(M_PI_4/2);
		static const double EAST = 15*(M_PI_4/2);
		
	};
	

	/**
	 * This is the base class for all EdgeDetectors which are made by using two, X and Y masks.
	 * The base operator are defined in the class and the base operators like Sobel, Prewitt and
	 * Roberts.
	 **/
	template <class T, class U=bool> class DifferentialEdgeDetector: public EdgeDetector<T,U>
	{
	public:
		/**
		 * Constructor which will be used when implemented mask are used.
		 *
		 * @param mask The implemented mask used in EdgeDetection.
		 * @param radius The radius used in suppresNonMaxima. Note that when
		 *               radius is zero suppression isn't made.
		 * @param borderAction How to handel the borders when needeed.
		 **/
		DifferentialEdgeDetector(EdgeMask mask, int radius=0, BorderAction borderAction = BORDER_REFLECT):
			EdgeDetector<T,U>(radius,borderAction),_Mask(mask)
		{ setMask(mask);}

		/**
		 * Constructor which will be used when own mask are used.
		 *
		 * @param maskX The convolution mask which is used calculatin the X gradient.
		 * @param maskY The convolution mask which is used calculatin the Y gradient.
		 * @param radius The radius used in suppresNonMaxima. Note that when
		 *               radius is zero suppression isn't made.
		 * @param borderAction How to handel the borders when needeed.
		 **/
		DifferentialEdgeDetector(ConvolutionMask<double>& maskX, ConvolutionMask<double>& maskY, int radius=0, BorderAction borderAction = BORDER_REFLECT):
			EdgeDetector<T,U>(radius,borderAction), _ConvolutionMaskX(maskX), _ConvolutionMaskY(maskY), _Mask(SOBEL) {}

		/**
		 * The Destructor of Differentian edge detector
		 **/
		~DifferentialEdgeDetector(){}

		/**
		 * Function set's the mask (enumeration Mask) for EdgeDetector.
		 *
		 * @param mask The Mask wanted to use.
		 **/ 
		void setMask(EdgeMask mask);

		/**
		 * Get Convolution mask gives the convolution mask.
		 *
		 * @param gradient The X or Y convolution mask.
		 **/
		ConvolutionMask<double> getConvolutionMask(Gradient gradient);
		
		/**
		 * Makes the gradient X and Y matrices from matrix given in parameter. 
		 *
		 * @param mat util::Matrix which wanted to have Gradients.
		 **/
		util::Matrix<Point<double> > getGradientMatrices(const util::Matrix<T>& mat);
		
		/**
		 * Makes the operation for the matrix given in parameter.
		 * NOTE that the same Border handling is used everywhere.
		 *
		 * @param mat the util::Matrix which wanted to use sobel.
		 * @return the new sobel matrix which indicates the edges.
		 *         type of matrix is spesified on template parameter U.
		 **/
		util::Matrix<U> getTransformedImage(const util::Matrix<T>& mat) throw (ImageTransformException&);

	protected:
	/**
		 * The constructor for the child classes.
		 **/
		DifferentialEdgeDetector(int radius, BorderAction borderAction):
			EdgeDetector<T,U>(radius,borderAction), _ConvolutionMaskX(), _ConvolutionMaskY(),_Mask(SOBEL) {}
		/**
		 * The Convolution mask for the X and Y gradient.
		 **/
		ConvolutionMask<double> _ConvolutionMaskX;
		ConvolutionMask<double> _ConvolutionMaskY;
		
	private:
		EdgeMask _Mask;
	};
	
	/**
	 * This Class makes the Canny edge detection. It uses the gaussian mean and
	 * derivative mask to make the result. First the convolution is made for
	 * the matrix with gaussian mask then the suppress non maxima is made if
	 * the supressRadius is not zero and then usally the Hysteresis threshold is
	 * used when using Canny operator therefore you should use Hysteresis threshold
	 * for the matrix which you get from the getTransformedImage.
	 **/
	template <class T, class U=bool> class Canny: public DifferentialEdgeDetector<T,U>
	{
	public:
		/**
		 * This constructor makes the Canny masks. 
		 *
		 * @param sigma The standard deviation at the gausian masks.
		 * @param supressRadius The radius used in suppresNonMaxima. Note that when
		 *               radius is zero suppression isn't made.
		 * @param borderAction How to handel the borders when needeed.
		 **/
		Canny(double sigma=2.0, int suppressRadius=0, BorderAction borderAction = BORDER_REFLECT);
		/**
		 * The destructor of Canny.
		 **/
		~Canny(){}
		
		/**
		 * Get the standard deviation.
		 **/
		double getSigma(){return _dSigma;}
		/**
		 * Set the standard deviation. When the sigma is set the gaussian masks are
		 * calculated again.
		 **/
		void setSigma(double sigma);
			
	private:
		/**
		 * This fuction makes the gaussian mask used in the Canny operator.
		 *
		 * @param sigma The standard deviation in the gaussian mask.
		 **/
		ConvolutionMask<double> makeCannyMask(double sigma);
		/**
		 * The standard deviation in the gaussian mask.
		 **/
		double _dSigma;
	};
			
	/**
	 * Laplace of gaussian detects edges in a local 3-by-3 neighborhood
	 * using a simple convolution mask:
	 * <pre>
	 *   0 -1  0
	 *  -1  4 -1
	 *   0 -1  0
	 * </pre>
	 **/
	template <class T, class U=bool> class LaplaceOfGaussian: public EdgeDetector<T,U>
	{	
	public:
		/**
		 * Constructor for LaplaceOfGaussian operator.
		 *
		 * @param radius The radius how width the supress non maxima will be done
		 *               Note if zero given suppresNonMaxima is not made.
		 * @param borderAction for the border handling. 
		 **/
		 LaplaceOfGaussian(int radius=0, BorderAction borderAction = BORDER_CROP):
			EdgeDetector<T,U>(radius,borderAction){}

		/**
		 * The destructor of LaplaceOfGaussian.
		 */
		~LaplaceOfGaussian(){}
		
		/**
		 * Makes the LaplaceOfGaussian operation for the matrix given in parameter.
		 *
		 * @param mat an input image
		 * @return the new LOG matrix which indicates the edges.
		 *         The type of the matrix is specified by the template parameter U.
		 **/
		util::Matrix<U> getTransformedImage(const util::Matrix<T>& mat) throw (ImageTransformException&);

	};

	template <class T, class U> util::Matrix<T> EdgeDetector<T,U>::suppressNonMaxima(const util::Matrix<T>& matMagnitudes,
																																									 const util::Matrix<T>& matAngle,
																																									 int radius, util::MatrixUtils::ExtendType extendType)
			throw (EdgeDetectorException&)
	{
		// if the radius is zero then the supress nonMaxima is not made.
		if(radius == 0)return matMagnitudes;
		
		util::Matrix<T> matMagnitude(matMagnitudes);
		matMagnitude = util::MatrixUtils::extend(matMagnitude,radius,extendType);
		util::Matrix<T> result(matMagnitudes);
		T magnitude,angle;		

		for(int i=radius,r=0;i<matMagnitude.getRows()-radius;i++,r++)
			for(int j=radius,c=0;j<matMagnitude.getColumns()-radius;j++,c++)
				{
					magnitude = matMagnitude(i,j);
					angle = matAngle(r,c);

					if(angle < WEST)
						{
							if(angle < NORTH)
								{
									if(angle < NORTH_EAST)
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i, j+k) > magnitude || matMagnitude(i, j-k) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}//NORTH_EAST
									else
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i+k, j+k) > magnitude || matMagnitude(i-k, j-k) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}//else NORTH_EAST
								}//NORTH
							else
								{
									if(angle < NORTH_WEST)
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i+k, j) > magnitude || matMagnitude(i-k, j) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}//NORTH_WEST
									else
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i+k, j-k) > magnitude || matMagnitude(i-k, j+k) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}// else NORTH_WEST
								}// else NORTH
						}//WEST
					else
						{
							if(angle < SOUTH)
								{
									if(angle < SOUTH_WEST)
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i, j-k) > magnitude || matMagnitude(i, j+k) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}//SOUTH_WEST
									else
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i-k, j-k) > magnitude || matMagnitude(i+k, j+k) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}//else SOUTH_WEST
								}//SOUTH
							else
								{
									if(angle < SOUTH_EAST)
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i-k, j) > magnitude || matMagnitude(i+k, j) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}//SOUTH_EAST
									else if(angle < EAST)
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i-k, j+k) > magnitude || matMagnitude(i+k, j-k) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}//else if SOUTH
									else
										{
											for(int k=1;k<=radius;k++)
												if(matMagnitude(i, j+k) > magnitude || matMagnitude(i, j-k) >= magnitude)
													{
														result(r,c) = 0;
														break;
													}
										}//else SOUTH_EAST
								}//else SOUTH
						}//else WEST
				}//for

		return result;
	}// suppressNonMaxima

	
	template <class T, class U>	util::Matrix<double> EdgeDetector<T,U>::getMagnitude(const util::Matrix<Point<T> >& mat)
		throw (EdgeDetectorException&)
	{
		int rows = mat.getRows();
		int columns = mat.getColumns();
		
		util::Matrix<double> result(rows, columns);

		// calculate the absolute value for both and then sum the values
		const Point<T>* matData = mat.getData();
		double* matRData = result.getData();
		
		for(int i=0;i<rows*columns;i++,matRData++,matData++)
			*matRData = double(sqrt(pow(double(matData->x),2)+pow(double(matData->y),2)));
		
		return result;
	}		

	template <class T, class U>	util::Matrix<double> EdgeDetector<T,U>::getAngle(const util::Matrix<Point<T> >& mat)
		throw (EdgeDetectorException&)
	{
		int rows = mat.getRows();
		int columns = mat.getColumns();
		
		util::Matrix<double> result(rows, columns);
		
		// calculate the absolute value for both and then sum the values
		const Point<T>* matData = mat.getData();
		double* matRData = result.getData();

		// remember to check that divider is not zero
		for(int i=0;i<rows*columns;i++,matRData++,matData++)
			{ // check if the divider id zero
				if(matData->x == T(0))
					{
						// check if grater than zero set the value to pi/2
						if(matData->y > T(0))*matRData = M_PI_2;
						// check if the value is less than zero set value to 3pi/2
						else if(matData->y < T(0))*matRData = 3*M_PI_2;
						// no need to use else beacause pixel's value is zero
						// allready
					}
				else
					{
						*matRData = atan(double(matData->y)/double(matData->x));
						// all angles is in range 0 - 2pi so check the values
						if(*matRData < T(0))*matRData += 2*M_PI;
					}
			}
		
		return result;
	}

	template <class T, class U> util::Matrix<T> EdgeDetector<T,U>::getGradient(const util::Matrix<Point<T> >& mat, Gradient gradient)
		throw (EdgeDetectorException&)
	{
		int rows = mat.getRows();
		int cols = mat.getColumns();
		const Point<T>* matData = mat.getData();
		util::Matrix<T> result(rows,cols);
		T* resultData = result.getData();

		if(gradient == GRADIENT_X)
			for(int i=0;i<rows*cols;i++,matData++,resultData++)
				*resultData = U(matData->x);
		else 
			for(int i=0;i<rows*cols;i++,matData++,resultData++)
				*resultData = U(matData->y);

		return result;
	}

	template <class T, class U> void EdgeDetector<T,U>::setGradient(util::Matrix<Point<T> >& matGradient,
																																	const util::Matrix<T>& mat, Gradient gradient)
			throw (EdgeDetectorException&)
	{
		int rows = matGradient.getRows();
		int cols = matGradient.getColumns();
		if(mat.getRows() != rows || mat.getColumns() != cols)
			throw EdgeDetectorException("EdgeDetector<T,U>::setGradient(util::Matrix<Point<T> >&, const util::Matrix<T>&, Gradient ): Matrix sizes differ.");
		
		Point<T>* matGradData = matGradient.getData();
		const T* matData = mat.getData();

		if(gradient == GRADIENT_X)
			for(int i=0;i<rows*cols;i++,matData++,matGradData++)
				matGradData->x = *matData;
		else 
			for(int i=0;i<rows*cols;i++,matData++,matGradData++)
				matGradData->y = *matData;
	}

	
	template<class T,class U> void DifferentialEdgeDetector<T,U>::setMask(EdgeMask mask)
	{
		if(mask == SOBEL)
			{
				double maskFactorsX[] = {-1,-2,-1,0,0,0,1,2,1}; // the sobel mask factors
				double maskFactorsY[] = {-1,0,1,-2,0,2,-1,0,1}; // the sobel mask factors
				ConvolutionMask<double> maskX(3,3,maskFactorsX); // the X mask
				ConvolutionMask<double> maskY(3,3,maskFactorsY); // the Y mask
				_ConvolutionMaskX = maskX;
				_ConvolutionMaskY = maskY;
			}
		else if(mask == PREWITT)
			{
				double maskFactorsX[] = {-1,-1,-1,0,0,0,1,1,1}; // the prewitt mask factors
				double maskFactorsY[] = {-1,0,1,-1,0,1,-1,0,1}; // the prewitt mask factors
				ConvolutionMask<double> maskX(3,3,maskFactorsX); // the X mask
				ConvolutionMask<double> maskY(3,3,maskFactorsY); // the Y mask
				_ConvolutionMaskX = maskX;
				_ConvolutionMaskY = maskY;
			}
		else if(mask == ROBERTS)
			{
				double maskFactorsX[] = {1,0,0,-1}; // the roberts mask factors
				double maskFactorsY[] = {0,1,-1,0}; // the roberts mask factors
				ConvolutionMask<double> maskX(2,2,maskFactorsX); // the X mask
				ConvolutionMask<double> maskY(2,2,maskFactorsY); // the Y mask
				_ConvolutionMaskX = maskX;
				_ConvolutionMaskY = maskY;
			}
	}

	template<class T,class U> ConvolutionMask<double> DifferentialEdgeDetector<T,U>::getConvolutionMask(Gradient gradient)
	{
		if(gradient == GRADIENT_X) return _ConvolutionMaskX;
		return _ConvolutionMaskY;
	}
	
	
	template <class T, class U> util::Matrix<Point<double> > DifferentialEdgeDetector<T,U>::getGradientMatrices(const util::Matrix<T>& mat)
	{
		int rows=mat.getRows();
		int cols=mat.getColumns();
		
		util::Matrix<T> tempX(Image::convolve(mat,_ConvolutionMaskX, _BorderAction));
		util::Matrix<T> tempY(Image::convolve(mat,_ConvolutionMaskY, _BorderAction));
		
		util::Matrix<Point<double> > result(rows,cols);
		T* matXData = tempX.getData();
		T* matYData = tempY.getData();
		Point<double>* resultData = result.getData();

		for(int i=0;i<rows*cols;i++,resultData++,matXData++,matYData++)
			{
				resultData->x = double(*matXData);
				resultData->y = double(*matYData);
			}
		
		return result;
	}
	
	template <class T, class U> util::Matrix<U> DifferentialEdgeDetector<T,U>::getTransformedImage(const util::Matrix<T>& mat)
		throw (ImageTransformException&)
	{
		// first calculate the gradient matrices
		util::Matrix<Point<double> > gradients(getGradientMatrices(mat));

		// then calculate the magnitude and angle matrices
		util::Matrix<double> matMagnitude(getMagnitude(gradients));
		util::Matrix<double> matAngle(getAngle(gradients));
		
		// and at the end lets make the suppressNonMaxima
		util::Matrix<double> suppress(suppressNonMaxima(matMagnitude,matAngle,_iradius));
		
		// then before returning typycasting all items to U
		return util::Matrix<U>(suppress);
	}

	template <class T, class U> util::Matrix<U> LaplaceOfGaussian<T,U>::getTransformedImage(const util::Matrix<T>& mat) throw (ImageTransformException&)
	{
		double maskFactors[] = {0,-1,0,-1,4,-1,0,-1,0}; // the LOG mask factors
		ConvolutionMask<double> LOGMask(3,3,maskFactors); // the LOG mask

		// then before returning typycast all items to U
		return util::Matrix<U>(Image::convolve(mat,LOGMask, _BorderAction));
	}

	template <class T,class U> Canny<T,U>::Canny(double sigma, int suppressRadius, BorderAction borderAction):
		DifferentialEdgeDetector<T,U>(suppressRadius,borderAction),_dSigma(sigma)
	{
		_ConvolutionMaskX = makeCannyMask(_dSigma);
		ConvolutionMask<double> tmp(_ConvolutionMaskX.getColumns(),_ConvolutionMaskX.getRows(),_ConvolutionMaskX.getData());
		_ConvolutionMaskY = tmp;
	}

	template <class T,class U> void Canny<T,U>::setSigma(double sigma)
	{
		_dSigma = sigma;
		_ConvolutionMaskX = makeCannyMask(_dSigma);
		ConvolutionMask<double> tmp(_ConvolutionMaskX.getColumns(),_ConvolutionMaskX.getRows(),_ConvolutionMaskX.getData());
		_ConvolutionMaskY = tmp;
	}
	
	template <class T,class U> ConvolutionMask<double> Canny<T,U>::makeCannyMask(double sigma)
	{
		// first make the meanGaussian mask
		int maskRadius=30;
		double squareSigma = sigma*sigma;
		double square2Sigma = 2.0*squareSigma;
		double divider = 6.0*M_PI*squareSigma;
		double i=-double(maskRadius);
		int j=0;

		// find the right size.
		do
			{
				if(exp(-i*i/square2Sigma)>0.0001)break;
				else j++;
				i++;
			}while(i<=maskRadius);

		// check all values which are bigger than 0.0001
		util::List<double> mFactors;
		for(double d=-maskRadius+j;d<=maskRadius-j;d++)
			{		
				double j=d+0.5,k=d-0.5;
				double tmp = (exp(-d*d/square2Sigma)+
											exp(-j*j/square2Sigma)+
											exp(-k*k/square2Sigma)
											)/divider;
				mFactors += tmp;
			}

		int mLen = mFactors.getLength();
		// then make the derivate Gaussian mask
		util::List<double> dFactors;
		for(double j=-double(mLen>>1);j<=mLen>>1;j++)
			dFactors += -j/squareSigma * exp(-j*j/square2Sigma);
		
		// then make the convolution for the factors
		util::List<double> factors(dsp::Convolution::conv(mFactors, dFactors, true));
		// then the convolution mask
		ConvolutionMask<double> filter(1,factors.getLength(),factors.getData());
				
		return filter;
	}
}}
#endif
