/*********************************************************************
 * This file is part of the cpplibs suite.
 *
 * Copyright (C) 2001 Topi Menp
 * All rights reserved.
 *
 * This program is free software. You can redistribute and/or modify
 * it under the terms of the free software licence found in the
 * accompanying file "COPYING". The licence terms must always be
 * redistributed with this source file. The above copyright notice
 * must be reproduced in all modified and unmodified copies of this
 * source file.
 *
 * $Revision: 1.6 $
 *********************************************************************/

#include "List.h"
#include "Number.h"

using namespace std;

namespace util
{
	typedef List<SmartPtr<Computable> > CompList;
	
	template <> void List<std::string>::addElements(int count,...)
	{
		if (_iCurrentItems + count > _iSize)
			{
				int size = _iSize;
				while (_iCurrentItems + count > size) size += _iBlockSize;
				enlargeArray(size);
			}
		va_list argp;
		// initalize var ptr
		va_start(argp, count); 

		// repeat for each arg
		while (count--)
			addElement(std::string(va_arg(argp,const char*)));

		// done with args
		va_end(argp);
	}

	template <> std::string CompList::toString() const
	{
		string result="(";
		if (_iCurrentItems > 0) result.append(_internalArray[0]->toString());
		for (int i=1;i<_iCurrentItems;i++)
			result.append(","+_internalArray[i]->toString());
		return result + ")";
	}

	//************************************************************
	//Calculation operations for computable objects.
	//************************************************************
	
	template <> SmartPtr<Computable> CompList::plus(const Computable& c, bool switched) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::plus(const Computable&,bool): List lengths differ.");
				
				for (int i=0;i<_iCurrentItems;i++)
					{
						if (!switched)
							result->addElement(_internalArray[i] + lst->_internalArray[i]);
						else				
							result->addElement(lst->_internalArray[i] + _internalArray[i]);
					}
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					{
						if (!switched)
							result->addElement(*_internalArray[i] + c);
						else				
							result->addElement(c + *_internalArray[i]);
					}
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::minus(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::minus(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] - *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] - c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::multiply(const Computable& c, bool switched) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::multiply(const Computable&,bool): List lengths differ.");
				
				for (int i=0;i<_iCurrentItems;i++)
					{
						if (!switched)
							result->addElement(*_internalArray[i] * *lst->_internalArray[i]);
						else				
							result->addElement(*lst->_internalArray[i] * *_internalArray[i]);
					}
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					{
						if (!switched)
							result->addElement(*_internalArray[i] * c);
						else				
							result->addElement(c * *_internalArray[i]);
					}
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::divide(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::divide(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] / *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] / c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::modulus(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::modulus(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] % *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] % c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::arrayIndex(const Computable& c) const
		throw (ComputationException&)
	{
		const Number* index = dynamic_cast<const Number*>(&c);
		if (index)
			{
				int i = index->intValue();
				if (i < 0 || i >= _iCurrentItems)
					throw ComputationException("List::arrayIndexconst Computable&): array index out of bounds.");
				//Return the value at index, keep ownership of the pointer in list
				SmartPtr<Computable> result(_internalArray[i].get());
				result.release();
				return result;
			}
		throw OperationNotDefinedException("List::arrayIndex(const Computable&): List[" + c.getClassName() + "] is not defined.");
	}

	template <> SmartPtr<Computable> CompList::functionCall(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		for (int i=0;i<_iCurrentItems;i++)
			result->addElement(_internalArray[i]->functionCall(c));
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::negate() const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		for (int i=0;i<_iCurrentItems;i++)
			result->addElement(_internalArray[i]->negate());
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::complement() const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		for (int i=0;i<_iCurrentItems;i++)
			result->addElement(_internalArray[i]->complement());
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::less(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::less(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] < *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] < c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::greater(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::greater(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] > *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] > c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::lessOrEqual(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::lessOrEqual(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] <= *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] <= c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::greaterOrEqual(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::greaterOrEqual(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] >= *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] >= c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::equal(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::equal(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] == *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] == c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::different(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::different(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] != *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] != c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::logAnd(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::logAnd(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] && *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] && c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::logOr(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::logOr(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] || *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] || c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::binAnd(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::binAnd(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] & *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] & c);
			}
		return SmartPtr<Computable>(result.release());
	}

	template <> SmartPtr<Computable> CompList::binOr(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::binOr(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] | *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] | c);
			}
		return SmartPtr<Computable>(result.release());
	}
	
	template <> SmartPtr<Computable> CompList::binXor(const Computable& c) const
		throw (ComputationException&)
	{
		SmartPtr<CompList> result(new CompList(_iCurrentItems));
		const CompList* lst = dynamic_cast<const CompList*>(&c);
		if (lst)
			{
				if (lst->_iCurrentItems != _iCurrentItems)
					throw ComputationException("List::binXor(const Computable&): List lengths differ.");

				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] ^ *lst->_internalArray[i]);
			}
		else
			{
				for (int i=0;i<_iCurrentItems;i++)
					result->addElement(*_internalArray[i] ^ c);
			}
		return SmartPtr<Computable>(result.release());
	}
	
}
